"""Top-level package for skippa.

The pipeline module defines the main Skippa methods.
The transformers subpackage contains various transformers used in the pipeline.
"""

__author__ = """Robert van Straalen"""
__email__ = 'tech@datasciencelab.nl'

from .pipeline import Skippa, columns
